from cStringIO import StringIO
import itertools
import maya
import mimetools
import mimetypes
import os
import shutil
import tempfile
import urllib
import urllib2
import zipfile


ASSETS_DIR_NAME = r'assets'
UPLOAD_URL = r'http://renderbuzz.com/api/upload/'


class MultiPartForm(object):
    """Accumulate the data to be used when posting a form."""

    def __init__(self):
        self.form_fields = []
        self.files = []
        self.boundary = mimetools.choose_boundary()
        return
    
    def get_content_type(self):
        return 'multipart/form-data; boundary=%s' % self.boundary

    def add_field(self, name, value):
        """Add a simple field to the form data."""
        self.form_fields.append((name, value))
        return

    def add_file(self, fieldname, filename, fileHandle, mimetype=None):
        """Add a file to be uploaded."""
        body = fileHandle.read()
        if mimetype is None:
            mimetype = mimetypes.guess_type(filename)[0] or 'application/octet-stream'
        self.files.append((fieldname, filename, mimetype, body))
        return
    
    def __str__(self):
        """Return a string representing the form data, including attached files."""
        # Build a list of lists, each containing "lines" of the
        # request.  Each part is separated by a boundary string.
        # Once the list is built, return a string where each
        # line is separated by '\r\n'.  
        parts = []
        part_boundary = '--' + self.boundary
        
        # Add the form fields
        parts.extend(
            [ part_boundary,
              'Content-Disposition: form-data; name="%s"' % name,
              '',
              value,
            ]
            for name, value in self.form_fields
            )
        
        # Add the files to upload
        parts.extend(
            [ part_boundary,
              'Content-Disposition: file; name="%s"; filename="%s"' % \
                 (field_name, filename),
              'Content-Type: %s' % content_type,
              '',
              body,
            ]
            for field_name, filename, content_type, body in self.files
            )
        
        # Flatten the list and add closing boundary marker,
        # then return CR+LF separated data
        flattened = list(itertools.chain(*parts))
        flattened.append('--' + self.boundary + '--')
        flattened.append('')
        return '\r\n'.join(flattened)


class RBUploader(object):
    
    def __init__(self, username, password, filepath):
        self.url = UPLOAD_URL
        self.username = username
        self.password = password
        self.filepath = filepath
        self.filename = os.path.basename(filepath)
        
    def _get_form(self):
        form = MultiPartForm()
        form.add_field('username', self.username)
        form.add_field('password', self.password)
        with open(self.filepath, 'rb') as f:
            form.add_file('file', self.filename, f)
        return form
        
    def _get_request(self):
        form = self._get_form()
        body = str(form)
        request = urllib2.Request(self.url)
        request.add_header('Content-type', form.get_content_type())
        request.add_header('Content-length', len(body))
        request.add_data(body)
        return request
        
    def get_response(self):
        request = self._get_request()
        try:
            urllib2.urlopen(request).read()
        except urllib2.HTTPError as e:
            msg = e.read()
            raise Exception(msg)


class RBUtils(object):
    
    @staticmethod
    def get_temp_dir():
        return tempfile.mkdtemp(prefix='rb_')
        
    @staticmethod
    def get_assets_dir(base_dir):
        assets_dir = os.path.join(base_dir, ASSETS_DIR_NAME)
        if not os.path.exists(assets_dir):
            os.makedirs(assets_dir)
        return assets_dir
        
    @staticmethod
    def copy_files(paths_list, dst):
        for path in paths_list:
            if os.path.exists(path):
                shutil.copy(path, dst)
                
    @staticmethod
    def list_files(dir):
        result = []
        rel_root = os.path.abspath(os.path.join(dir, os.pardir))
        for root, dirs, files in os.walk(dir):
            for file in files:
                filename = os.path.join(root, file)
                result.append(filename)
        return result
        
    @staticmethod
    def zip(files, target_dir, archive_name):
        zipf = None
        os.chdir(target_dir)
        archive_name = '{0}.zip'.format(archive_name)
        with zipfile.ZipFile(archive_name, 'w', zipfile.ZIP_DEFLATED, allowZip64=True) as zipf:
            for f in files:
                arc_name = os.path.join(os.path.relpath(os.path.dirname(f), target_dir), os.path.basename(f))
                zipf.write(f, arc_name)
                
        return os.path.join(target_dir, archive_name)
        
    @staticmethod
    def zip_archive(dir):
        scene_files = RBUtils.list_files(dir)
        scene_name = MayaUtils.get_scene_name()
        archive_filepath = RBUtils.zip(scene_files, dir, scene_name)
        return archive_filepath
        
    @staticmethod
    def upload_file(username, password, filepath):
        uploader = RBUploader(username, password, filepath)
        uploader.get_response()


class MayaUtils(object):
    
    @staticmethod
    def get_assets():
        return [maya.cmds.getAttr(f + '.fileTextureName') for f in maya.cmds.ls(type='file')]
        
    @staticmethod        
    def get_scene_filepath():
        return maya.cmds.file(query=True, sceneName=True)
        
    @classmethod
    def get_scene_filename(cls):
        return os.path.basename(cls.get_scene_filepath())

    @classmethod
    def get_scene_name(cls):
        name, _ = os.path.splitext(cls.get_scene_filename())
        return name
        
    @staticmethod
    def load_scene(path):
        return maya.cmds.file(path, open=True)
        
    @staticmethod
    def save_scene():
        return maya.cmds.file(save=True)
        
    @classmethod
    def copy_scene(cls, dir):
        filename = cls.get_scene_filename()
        filepath = os.path.join(dir, filename)
        maya.cmds.file(rename=filepath)
        return cls.save_scene()
        
    @staticmethod
    def get_workspace_filepath():
        return maya.cmds.workspace(query=True, dir=True)
        
    @staticmethod
    def load_workspace(path):
        return maya.cmds.workspace(path, openWorkspace=True)
        
    @staticmethod
    def repath_assets():
        for asset in maya.cmds.ls(type='file'):
            attr = asset + '.fileTextureName'
            filepath = maya.cmds.getAttr(attr)
            filename = os.path.basename(filepath)
            value = os.path.join(os.curdir, ASSETS_DIR_NAME, filename)
            maya.cmds.setAttr(attr, value, type="string")

    
class Scene(object):

    def __enter__(self):
        self._user_scene_path = MayaUtils.get_scene_filepath()
        self._user_workspace_path = MayaUtils.get_workspace_filepath()
        self.temp_dir = RBUtils.get_temp_dir()
        MayaUtils.load_workspace(self.temp_dir)
        return self
        
    def archive(self):
        MayaUtils.copy_scene(self.temp_dir)
        
        assets_dir = RBUtils.get_assets_dir(self.temp_dir)
        assets_filepaths = MayaUtils.get_assets()            
        RBUtils.copy_files(assets_filepaths, assets_dir)
        
        MayaUtils.repath_assets()
        MayaUtils.save_scene()
        return self.temp_dir
                
    def __exit__(self, *args, **kwargs):
        MayaUtils.load_scene(self._user_scene_path)
        MayaUtils.load_workspace(self._user_workspace_path)


class RBUploaderGUI(object):
    GUI_ID = "RB_GUI"
    
    def __init__(self):
        if maya.cmds.window(self.GUI_ID, exists=True):
            maya.cmds.deleteUI(self.GUI_ID)
            
        window = maya.cmds.window(self.GUI_ID, title="RenderBuzz Uploader", 
                                  wh=(270, 280), sizeable=False, resizeToFitChildren=True,
                                  minimizeButton=False, maximizeButton=False)
        layout = maya.cmds.columnLayout(columnAttach=('both', 5), columnAlign='left', rowSpacing=5)
        
        maya.cmds.separator(style='none')    
        pref_dir = maya.cmds.internalVar(userPrefDir=True)
        logo_filepath = os.path.join(pref_dir, 'scripts', 'renderbuzz', 'rblogo.png')
        maya.cmds.image(w=260, h=50, image=logo_filepath)
        
        maya.cmds.text(label="E-mail:")
        self.username_txt_box = maya.cmds.textField('usernameTxtBox', w=260)
        
        maya.cmds.text(label="Password:")
        self.password_txt_box = maya.cmds.textField('passwordTxtBox', w=260)
        
        maya.cmds.separator(h=50, style='none')    
        self.upload_btn = maya.cmds.button(label="Upload", w=260, h=50, command=self.upload_project)
    
        self.status_txt = maya.cmds.text("Welcome to RenderBuzz", w=260)
        
        maya.cmds.showWindow(window)

    @property        
    def username(self):
        username = maya.cmds.textField(self.username_txt_box, q=True, text=True)
        return username.encode('ascii', 'replace')
        
    @property
    def password(self):
        password = maya.cmds.textField(self.password_txt_box, q=True, text=True)
        return password.encode('ascii', 'replace')
        
    def change_status(self, txt):
        maya.cmds.text(self.status_txt, edit=True, label=txt)
        maya.cmds.refresh(currentView=True, force=True)
    
    def upload_project(self, *args):        
        username = self.username
        password = self.password
        
        try:
            maya.cmds.waitCursor(state=True)
            maya.cmds.button(self.upload_btn, edit=True, enable=False)
        
            self.change_status("Preparing files")
            with Scene() as scene:
                archive_dir = scene.archive()
            
            self.change_status("Creating archive")
            zip_filepath = RBUtils.zip_archive(archive_dir)
        
            self.change_status("Uploading")
            RBUtils.upload_file(username, password, zip_filepath)
        except Exception as e:
            self.change_status(str(e))        
        else:
            self.change_status("You can now log in to RenderBuzz and start rendering")        
        finally:
            maya.cmds.waitCursor(state=False)
            maya.cmds.button(self.upload_btn, edit=True, enable=True)


def main():    
    RBUploaderGUI()


if __name__ == '__main__':
    main()